/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/

#include <stdio.h>
#include "oo7.h"
#include "globals.h"

extern int overFlowCnt;

// methods for association and association iterator classes
// including the association overflow class

// constructor
Assoc::Assoc(collection<dbvoid>* cpt)
{
    int i;

    allocated =  BaseSize;  
    assocFile = cpt;  // collection 
    curSize = 0; 	// all entries are free 
    freePtr = 0;
    for (i=0;i<allocated;i++) members[i] = NULL;
    overflow = NULL;
}



void AssocBase::add(dbvoid* member) {
    int i;

    // see if any free spots remain in the pointer array 
    if (curSize < allocated) 
    {
	// at least one is available
	i = freePtr;
	while ((i < allocated) && (members[i] != NULL)) i++;
	if (i < allocated)
	{
	    members[i] = member;
	    curSize++;
	    freePtr = i+1;  // keep track of where we stopped 
	    return;
	}
    }

    // current array is full,
    // try all the overflow arrays (recursively)
    if (overflow != NULL) overflow->add(member);
    else {
	// add an overflow pointer array
	overFlowCnt++;
	overflow = new (*assocFile) AssocOverflow(assocFile);
	overflow->add(member);
    }
}
void AssocBase::remove(dbvoid* member) {
    register int i;

    // look first in pointer array
    if (curSize > 0)
    {
	// at least one pointer exists
	i = 0;
	for (i=0;i<allocated;i++)
	{
	    if (members[i] == member) {
		members[i] = NULL;
		if (i < freePtr) freePtr = i; // this entry is free now
		curSize--;
		return;
	    }
	}
    }
    // not in the pointer array try the overflow chain
    if (overflow != NULL) overflow->remove(member);
}


// constructor
AssocOverflow::AssocOverflow(collection<dbvoid>* cpt)
{
    int i;

    allocated =  OverflowSize;  
    assocFile = cpt;  // collection 
    curSize = 0; 	// all entries are free 
    freePtr = 0;	// keep track of first free entry
    for (i=0;i<allocated;i++) members[i] = NULL;
    overflow = NULL;
}






// methods for association iterator class

// constructor
AssocItr::AssocItr(Assoc*  assoc) {
    current = assoc;
    scanPtr = 0;
}

dbvoid* AssocItr::next() {
    int i;
    dbvoid *obj;

    while (1)
    {
	// first look in the current association block
	for (i = scanPtr; i < current->allocated; i++)
    	{
	    obj = current->member(i);
	    if (obj != NULL) 
	    {
		// found one!
		scanPtr = i+1;  // update scan pointer before returning
		return(obj);
	    }
	}
	// exhausted current pointer association block.  try to advance 
	// to next block
	if (current->overflow != NULL)
	{
	    current = current->overflow;
	    scanPtr = 0;
	}
	else return(NULL);
    }
}

AssocBase::~AssocBase()
// destructor
{
    if (overflow) delete overflow;  // recursively delete the overflow blocks
}
