/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/



////////////////////////////////////////////////////////////////////////////
//
// AtomicPart Methods
//
////////////////////////////////////////////////////////////////////////////

#include <libc.h>
#include <stdio.h>
#include "oo7.h"
#include "globals.h"
#include "VarParams.h"
#include "GenParams.h"
#include <sys/resource.h>
#include <E/sm_btree.h>

extern int	debugMode;
// this should maybe be in globals.h, but we want to avoid having
// to have templates everywhere.
persistent btree<int> BuildDateIdx;


extern char*	types[NumTypes];

////////////////////////////////////////////////////////////////////////////
//
// AtomicPart Constructor
//
////////////////////////////////////////////////////////////////////////////

AtomicPart::AtomicPart(dbint ptId, CompositePart *cp)
{
    if (debugMode) {
        printf("AtomicPart::AtomicPart(ptId = %d)\n", ptId);
    }

    // initialize internal state of new part
    id = ptId;
    int typeNo = (int) (random() % NumTypes);
    strncpy(type, types[typeNo], TypeSize);
    buildDate = MinAtomicDate +
        (int) (random() % (MaxAtomicDate-MinAtomicDate+1));
    x = (int) (random() % XYRange);
    y = (int) (random() % XYRange);
    // fill in a random document id (for query 9)
    docId = (int) random() % TotalCompParts + 1;

    // initialize space for connections to other parts
    from = new (CompPartSet) Assoc(&CompPartSet);
    to = new (CompPartSet) Assoc(&CompPartSet);
    partOf = cp;
    // AllParts.add(this);
    // BuildDateIdx.insert(buildDate,this); // done in CompositePart.e
	// explicitly now to facilitate bulk load
}

// destructor

AtomicPart::~AtomicPart()
{
    Connection* cn;
    AssocItr*	connI;
    AtomicPart* ap;

    // remove part from extent of all atomic parts
    // AllParts.remove (this);

    // establish iterator over connected parts in "to" direction
    connI = new AssocItr(to);
    cn = (Connection*) connI->next(); // get first connection record
    while (cn != NULL)
    {
         ap = cn->to;  // get the associated atomic part
	 ap->from->remove(cn); // remove back pointer to connection object
	 delete cn; // delete the connection object
	 cn = (Connection*) connI->next(); // get next connection record
    }
    delete connI;
    delete to; // delete the "to" associaton

    // establish iterator over connected parts in "from" direction
    connI = new AssocItr(from);
    cn = (Connection*) connI->next(); // get first connection record
    while (cn != NULL)
    {
         ap = cn->from;  // get the associated atomic part
	 ap->to->remove(cn); // remove back pointer to connection object
	 delete cn; // delete the connection object
	 cn = (Connection*) connI->next(); // get next connection record
    }
    delete connI;
    delete from; // delete the "from" associaton
    BuildDateIdx.remove(buildDate,this);
}


////////////////////////////////////////////////////////////////////////////
//
// AtomicPart swapXY Method for use in update traversals
//
////////////////////////////////////////////////////////////////////////////

void AtomicPart::swapXY()
{
    if (debugMode) {
        printf("                    AtomicPart::swapXY(ptId = %d, x = %d, y = %d)\n", id,x,y);
    }

    // exchange X and Y values

    int tmp = x;
    x = y;
    y = tmp;

    if (debugMode) {
        printf("                    [did swap, so x = %d, y = %d]\n", x, y);
    }
}


////////////////////////////////////////////////////////////////////////////
//
// AtomicPart toggleDate Method for use in update traversals
//
////////////////////////////////////////////////////////////////////////////

void AtomicPart::toggleDate()
{
    if (debugMode) {
        printf("                    AtomicPart::toggleDate(ptId = %d, buildDate = %d)\n",
	       id, buildDate);
    }

    // delete from index first
    BuildDateIdx.remove(buildDate,this);

    // increment build date if odd, decrement it if even

    if (buildDate % 2) {
	// odd case
	buildDate++;
    } else {
	// even case
	buildDate--;
    }

    // update the index
    BuildDateIdx.insert(buildDate,this);

    if (debugMode) {
        printf("                    [did toggle, so buildDate = %d]\n", buildDate);
    }
}


////////////////////////////////////////////////////////////////////////////
//
// AtomicPart DoNothing Method 
//
////////////////////////////////////////////////////////////////////////////

int RealWork; // set to one to make DoNothing do work.
int WorkAmount; // amount of work to do

void AtomicPart::DoNothing()
{
    struct timeval tv;
    struct timezone tz;
    int ignore;

    int i;

    if (id < 0) 
        printf("DoNothing: negative id.\n");

    if (debugMode) {
        printf("==> DoNothing(x = %d, y = %d, type = %s)\n", x, y, type);
    }

    if (RealWork) {
            for (i = 0; i < WorkAmount; i++) {
               gettimeofday(&tv, &tz);
            }
    }
}

