/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/

#include <libc.h>
#include <stdio.h>
#include <sys/resource.h>
#include <sys/time.h>

#include "oo7.h"
#include "globals.h"
#include "GenParams.h"
#include "VarParams.h"
#include "BenchParams.h"
#include "IntBtree.h"
#include "StringBtree.h"
#include "stats.h"
#include "stats_server.h"
#include "baidlist.h"

extern int RealWork; // set to one to make atomicpart::DoNothing
		// actually do some work
extern int WorkAmount;   // controls how much work DoNothings do

extern void SetParams(char* configFileName);

extern int  traverse7();
extern int  query1();
extern int  query2();
extern int  query3();
extern int  query4();
extern int  query5();
extern int  query6();
extern int  query7();
extern int  query8();
extern void insert1();
extern void delete1();
extern int  reorg1();
extern int  reorg2();
extern void Purge();

int overFlowCnt;

extern double ComputeWallClockTime(struct timeval *startWallTime, 
			           struct timeval *endWallTime);

extern double ComputeUserTime(struct rusage *startUsage, 
	        	      struct rusage *endUsage);

extern double ComputeSystemTime(struct rusage *startUsage, 
	 	                struct rusage *endUsage);


//////////////////////////////////////////////////////////////////
//
// Global Variables, etc., for Benchmarking ODB Operations
//
//////////////////////////////////////////////////////////////////


int	nextAtomicId=0; 
int	nextCompositeId=0;
int	nextComplexAssemblyId=0;
int	nextBaseAssemblyId=0;
int	nextModuleId = TotalModules;
int	debugMode = FALSE;

struct rusage startUsage, endUsage;
struct timeval startWallTime;
struct timeval endWallTime;
struct timezone ignoreTimeZone;
struct timeval startWarmTime;


char    *types[NumTypes] = {
		"type000", "type001", "type002", "type003", "type004",
		"type005", "type006", "type007", "type008", "type009"
	};

// here only to keep the make happy. used in gendb
BAIdList* private_cp;
BAIdList* shared_cp;

///////////////////////////////////////////////////////////////////////////
// ParseCommandLine parses the original shell call to "bench", determining 
// which operation to run, how many times to run it, and whether the
// individual runs are distinct transactions or are lumped together
// into one large transaction.
//////////////////////////////////////////////////////////////////////////
void ParseCommandLine(int argc, 
		      char**argv, 
		      int& opIndex,
		      int& repeatCount,
                      BenchmarkOp& whichOp,
	              int& manyXACTS) 
{
	printf("Call was: ");
	for (int foo = 0; foo < argc; foo++) {
		printf("%s  ", argv[foo]);
	}
	printf("\n");
	opIndex = 3;

	if (strcmp(argv[argc-1], "-d") == 0) debugMode = TRUE;
        else debugMode = FALSE;
        if (argc < 5)
        {
	    fprintf(stderr, "Usage: %s <configFileName> <count> <op> <xacts> [-d]\n", 
		argv[0]);
	    fprintf(stderr, "count = number of times to repeat op.\n");
	    fprintf(stderr, "Currently supported <op>s: t1, t1ww, t2a, t2b, t2c, t3a, t3b, t3c\n");
	    fprintf(stderr, "\t t4, t5do, t5undo, t6, t7, q1, q1ww, q2, q3, q4, q5, q6, q7, q8\n");
	    fprintf(stderr, "\t i, d, r1, r2, wu, m1, m2, m3, m4, m5, m6.\n");
	    fprintf(stderr, "<xacts> = single (all repetitions done in one xact)\n");
	    fprintf(stderr, "or many (each repetition a distinct xact.)\n");
	    exit(1);
	}

	if ((repeatCount = atoi(argv[2])) <= 0) 
	{
	    fprintf(stderr, "Usage: %s <configFileName> <count> <op> <xacts> [-d]\n", argv[0]);
	    exit(1);
        } 
	
	if (strcmp(argv[4], "many") == 0) {
            manyXACTS = 1;
        } 
	if (strcmp(argv[opIndex], "t1") == 0) {
	    whichOp = Trav1;
	} else if (strcmp(argv[opIndex], "t1ww") == 0) {
	    whichOp = Trav1WW;
	    if (debugMode == TRUE) WorkAmount = atoi(argv[argc-2]);
            else WorkAmount = atoi(argv[argc-1]);
            printf("WorkAmount = %d\n", WorkAmount);
	} else if (strcmp(argv[opIndex], "t2a") == 0) {
	    whichOp = Trav2a;
	} else if (strcmp(argv[opIndex], "t2b") == 0) {
	    whichOp = Trav2b;
	} else if (strcmp(argv[opIndex], "t2c") == 0) {
	    whichOp = Trav2c;
	} else if (strcmp(argv[opIndex], "t3a") == 0) {
	    whichOp = Trav3a;
	} else if (strcmp(argv[opIndex], "t3b") == 0) {
	    whichOp = Trav3b;
	} else if (strcmp(argv[opIndex], "t3c") == 0) {
	    whichOp = Trav3c;
	} else if (strcmp(argv[opIndex], "t4") == 0) {
	    whichOp = Trav4;
	} else if (strcmp(argv[opIndex], "t5do") == 0) {
	    whichOp = Trav5do;
	} else if (strcmp(argv[opIndex], "t5undo") == 0) {
	    whichOp = Trav5undo;
	} else if (strcmp(argv[opIndex], "t6") == 0) {
	    whichOp = Trav6;
	} else if (strcmp(argv[opIndex], "t7") == 0) {
	    whichOp = Trav7;
	} else if (strcmp(argv[opIndex], "t8") == 0) {
	    whichOp = Trav8;
	} else if (strcmp(argv[opIndex], "t9") == 0) {
	    whichOp = Trav9;
	} else if (strcmp(argv[opIndex], "t10") == 0) {
	    whichOp = Trav10;
	} else if (strcmp(argv[opIndex], "q1") == 0) {
	    whichOp = Query1;
	} else if (strcmp(argv[opIndex], "q2") == 0) {
	    whichOp = Query2;
	} else if (strcmp(argv[opIndex], "q3") == 0) {
	    whichOp = Query3;
	} else if (strcmp(argv[opIndex], "q4") == 0) {
	    whichOp = Query4;
	} else if (strcmp(argv[opIndex], "q5") == 0) {
	    whichOp = Query5;
	} else if (strcmp(argv[opIndex], "q6") == 0) {
	    whichOp = Query6;
	} else if (strcmp(argv[opIndex], "q7") == 0) {
	    whichOp = Query7;
	} else if (strcmp(argv[opIndex], "q8") == 0) {
	    whichOp = Query8;
	} else if (strcmp(argv[opIndex], "i") == 0) {
	    whichOp = Insert;
	} else if (strcmp(argv[opIndex], "d") == 0) {
	    whichOp = Delete;
	} else if (strcmp(argv[opIndex], "r1") == 0) {
	    whichOp = Reorg1;
	} else if (strcmp(argv[opIndex], "r2") == 0) {
	    whichOp = Reorg2;
// NEW
	} else if (strcmp(argv[opIndex], "wu") == 0) {
	    whichOp = WarmUpdate;
	} else {
	    fprintf(stderr, "ERROR: Illegal OO7 operation specified.\n");
	    fprintf(stderr, "Currently supported <op>s: t1, t1ww, t2a, t2b, t2c, t3a, t3b, t3c\n");
	    fprintf(stderr, "\t t4, t5do, t5undo, t6, t7, t8, t9, q1, q2, q3, q4, q5, q6, q7, q8\n");
	    fprintf(stderr, "\t i, d, r1, r2 wu\n");
	    exit(1);
        }
}                


main(int argc, char** argv)
{

	Module*  moduleH;
	char moduleName[40];
	ProcessStats    totalTime;
	ServerStats     totalSrvTime;
	char*	purgeVar;

        char  resultText[200];  // buffer to hold result of operation for
                                // printing outside of timing region.

	// initialize parameters for benchmark.
	SetParams(argv[1]);
	nextAtomicId  = TotalAtomicParts + 1;
	nextCompositeId = TotalCompParts + 1;

	// Compute structural info needed by the update operations,
        // since these operations need to know which id's should
        // be used next.

	int baseCnt = NumAssmPerAssm;
	int complexCnt = 1;
	for (int i = 1; i < NumAssmLevels-1; i++) {
            baseCnt = baseCnt * NumAssmPerAssm;
            complexCnt += complexCnt * NumAssmPerAssm;
	}
	nextBaseAssemblyId = TotalModules*baseCnt + 1;
	nextComplexAssemblyId = TotalModules*complexCnt + 1;
	nextAtomicId = TotalAtomicParts + 1;
	nextCompositeId = TotalCompParts + 1;

	int opIndex = 2;
	int repeatCount = 1;
	BenchmarkOp whichOp = Trav1;
        int manyXACTS = 0;

	// needed for insert and delete tests
	shared_cp = new BAIdList[TotalCompParts+NumNewCompParts+1];
	private_cp = new BAIdList[TotalCompParts+NumNewCompParts+1];


	// See if debug mode is desired, see which operation to run,
	// and how many times to run it.
        ParseCommandLine(argc, argv, opIndex, repeatCount, whichOp, manyXACTS); 

	purgeVar = getenv ("PURGE");
	if (purgeVar != NULL)
	{
		if (atoi(purgeVar) == 1)
		{
			E_BeginTransaction();
			printf("purginging db \n");
			Purge();
			printf("done purging db\n");
			E_CommitTransaction();
		}
		else printf("not purging db\n");
	}
	else printf("not purging db\n");

	totalTime.Start();
	totalSrvTime.Start();

        // Actually run the darn thing.
	for (int iter = 0; iter < repeatCount; iter++) 
	{
	    //////////////////////////////////////////////////////////////////
	    // Run an OO7 Benchmark Operation
	    //
	    //////////////////////////////////////////////////////////////////

	    printf("RUNNING OO7 BENCHMARK OPERATION %s, iteration = %d.\n", 
	           argv[opIndex], iter);

  	    // get wall clock time
            gettimeofday(&startWallTime, &ignoreTimeZone);

	    // get starting usage values.
	    getrusage(RUSAGE_SELF, &startUsage);

	    // Start a new transaction if either this is the first iteration
	    // of a multioperation transaction or we we are running each
	    // operate as a separate transaction

	    if ((iter == 0) || (manyXACTS)) E_BeginTransaction();

            // set random seed so "hot" runs are truly hot
            srandom(1);

	    // Use random module for the operation
            int moduleId = (int) (random() % TotalModules) + 1;
            sprintf(moduleName, "Module %08d", moduleId);
//	    printf("moduleName=%s\n",moduleName);
	    moduleH = (Module*) ModuleIdx->lookup(moduleName);
	    if (moduleH == NULL)
	    {
	        fprintf(stderr, "ERROR: Unable to access %s.\n", moduleName);
		E_AbortTransaction();
	        exit(1);
	    }

	    // Perform the requested operation on the chosen module
	    int count = 0;
	    int docCount = 0;
	    int charCount = 0;
	    int replaceCount = 0;

	    switch (whichOp) {
	        case Trav1:
	            count = moduleH->traverse(whichOp);
	            sprintf(resultText, "Traversal 1 DFS visited %d atomic parts.\n",
			             count);
  		    break;
		 case Trav1WW:
                    RealWork = 1;
                    whichOp = Trav1;  // so traverse methods don't complain
                    count = moduleH->traverse(whichOp);
                    whichOp = Trav1WW;  // for next (hot) traversal
                    sprintf(resultText, "Traversal 1WW DFS visited %d atomic parts.\n",
                                     count);
                    break;
	        case Trav2a:
	            count = moduleH->traverse(whichOp);
	            sprintf(resultText, "Traversal 2A swapped %d pairs of (X,Y) coordinates.\n",
 			         count);
		    break;
	        case Trav2b:
	            count = moduleH->traverse(whichOp);
	            sprintf(resultText, "Traversal 2B swapped %d pairs of (X,Y) coordinates.\n",
			             count);
		    break;
	        case Trav2c:
	            count = moduleH->traverse(whichOp);
	            sprintf(resultText, "Traversal 2C swapped %d pairs of (X,Y) coordinates.\n",
			             count);
		    break;
	        case Trav3a:
	            count = moduleH->traverse(whichOp);
	            sprintf(resultText, "Traversal 3A toggled %d dates.\n",
			             count);
		    break;
	        case Trav3b:
	            count = moduleH->traverse(whichOp);
	            sprintf(resultText, "Traversal 3B toggled %d dates.\n",
			             count);
		    break;
	        case Trav3c:
	            count = moduleH->traverse(whichOp);
	            sprintf(resultText, "Traversal 3C toggled %d dates.\n",
			            count);
		    break;
	        case Trav4:
	            count = moduleH->traverse(whichOp);
	            sprintf(resultText, "Traversal 4: %d instances of the character found\n",
			             count);
		    break;
	        case Trav5do:
	            count = moduleH->traverse(whichOp);
	            sprintf(resultText, "Traversal 5(DO): %d string replacements performed\n",
			             count);
		    break;
	        case Trav5undo:
	            count = moduleH->traverse(whichOp);
	            sprintf(resultText, "Traversal 5(UNDO): %d string replacements performed\n",
			         count);
		    break;
	        case Trav6:
	            count = moduleH->traverse(whichOp);
	            sprintf(resultText, "Traversal 6: visited %d atomic part roots.\n",
			             count);
		    break;
	        case Trav7:
	            count = traverse7();
		    sprintf(resultText, "Traversal 7: found %d assemblies using rand om atomic part.\n", 
			count);
		    break;
	        case Trav8:
	            count = moduleH->scanManual();
		    sprintf(resultText, "Traversal 8: found %d occurrences of character in manual.\n", 
			count);
		    break;
	        case Trav9:
	            count = moduleH->firstLast();
		    sprintf(resultText, "Traversal 9: match was %d.\n", 
			count);
		    break;

                case Trav10:
                    // run traversal #1 on every module.
                    count = 0;
                    whichOp = Trav1;  // so object methods don't complain
                    for (moduleId = 1; moduleId <= TotalModules; moduleId++) {
                        sprintf(moduleName, "Module %08d", moduleId);
	  	        moduleH = (Module*) ModuleIdx->lookup(moduleName);
   	                if (moduleH == NULL) {
                                fprintf(stderr,
                                        "ERROR: t10 Unable to access %s.\n",
                                         moduleName);
                                E_AbortTransaction();
                                exit(1);
                        }
                        count += moduleH->traverse(whichOp);
                    }
                    sprintf(resultText,
                           "Traversal 10 visited %d atomic parts in %d modules.\\n",
                                     count, TotalModules);
                    whichOp = Trav10;  // for next time around
                    break;           

	        case Query1:
	            count = query1();
	            sprintf(resultText, "Query one retrieved %d atomic parts.\n",
			             count);
		    break;
	        case Query2:
	            count = query2();
	            sprintf(resultText, "Query two retrieved %d qualifying atomic parts.\n",
			         count);
		    break;
	        case Query3:
	            count = query3();
	            sprintf(resultText, "Query three retrieved %d qualifying atomic parts.\n",
			         count);
		    break;
	        case Query4:
	            count = query4();
	            sprintf(resultText, "Query four retrieved %d (document, base assembly) pairs.\n",
			         count);
		    break;
	        case Query5:
	            count = query5();
	            sprintf(resultText, "Query five retrieved %d out-of-date base assemblies.\n",
			             count);
		    break;
	        case Query6:
	            count = query6();
	            sprintf(resultText, "Query six retrieved %d out-of-date assemblies.\n",
			         count);
		    break;
	        case Query7:
	            count = query7();
		    sprintf(resultText, "Query seven iterated through %d atomic part s.\n",
			             count);
		    break;
	        case Query8:
	            count = query8();
		    sprintf(resultText, "Query eight found %d atomic part/document m atches.\n",
			 count);
		    break;
	        case Insert:
	            insert1();
	            sprintf(resultText, "Inserted %d composite parts (a total of %d atomic parts.)\n",
		      NumNewCompParts, NumNewCompParts*NumAtomicPerComp);
		    break;
	        case Delete:
	            delete1();
	            sprintf(resultText, "Deleted %d composite parts (a total of %d atomic parts.)\n",
	             NumNewCompParts, NumNewCompParts*NumAtomicPerComp);
		    break;

		 case Reorg1:
		     count = reorg1();
		     sprintf(resultText, "Reorg1 replaced %d atomic parts.\n", 
			count);
		     break;

	    	 case Reorg2:
		     count = reorg2();
		     sprintf(resultText, "Reorg2 replaced %d atomic parts.\n", 
			count);
		     break;
// NEW
	        case WarmUpdate:
		    // first do the t1 traversal to warm the cache
	            count = moduleH->traverse(Trav1);
		    // then call T2 to do the update
	            count = moduleH->traverse(Trav2a);
	            sprintf(resultText, 
			"Warm update swapped %d pairs of (X,Y) coordinates.\n",
 			         count);
		     break;
	        default:
	            fprintf(stderr, "Sorry, that operation isn't available yet.\n");
		    E_AbortTransaction();
	            exit(1);
	    }

	    // Commit the current transaction if we are running the 
	    // last iteration or running a multitransaction test
	    if ((iter == repeatCount-1) || (manyXACTS)) E_CommitTransaction();

            // compute and report wall clock time
            gettimeofday(&endWallTime, &ignoreTimeZone);
	    printf("EXODUS, operation=%s, iteration=%d, elapsedTime=%f seconds\n",
               argv[opIndex], iter,
               ComputeWallClockTime(&startWallTime, &endWallTime));
            if (iter == 1) startWarmTime = startWallTime;

            // Compute and report CPU time.
	    getrusage(RUSAGE_SELF, &endUsage);
            fprintf(stdout, resultText);
	    fprintf(stdout, "CPU time: %f seconds.\n", 
	                ComputeUserTime(&startUsage, &endUsage) +
			ComputeSystemTime(&startUsage, &endUsage));
	    fprintf(stdout, "(%f seconds user, %f seconds system.)\n", 
	                ComputeUserTime(&startUsage, &endUsage),
			ComputeSystemTime(&startUsage, &endUsage));

	    if ((repeatCount > 2) && (iter == repeatCount-2)) 
	    {
	       // compute average hot time for 2nd through n-1 th iterations
               printf("EXODUS, operation=%s, average hot elapsedTime=%f seconds\n",
	       	  argv[opIndex], 
	          ComputeWallClockTime(&startWarmTime, &endWallTime)/(repeatCount-2)); 
	    }
	}

	//////////////////////////////////////////////////////////////////
	//
	// Shutdown 
	//
	//////////////////////////////////////////////////////////////////

	totalTime.Stop();
	totalSrvTime.Stop();
	fprintf(stdout, "Total stats (client,server):\n");
	totalTime.PrintStatsHeader(stdout);
	totalTime.PrintStats(stdout, "TotalCli");
	totalSrvTime.PrintStats(stdout, "TotalSrv");

	// Exit
	exit(0);
}

