/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/

#include <libc.h>
#include <stdio.h>

#include "oo7.h"
#include "globals.h"
#include "GenParams.h"
#include "BenchParams.h"

//////////////////////////////////////////////////////////////////
//
// Query #5 - find all base assemblies B that "use" a composite
// part with a more recent build date than B's build date.
//
/////////////////////////////////////////////////////////////////

extern int	debugMode;

//////////////////////////////////////////////////////////////////
//
// Routine to do Query #5
//
//////////////////////////////////////////////////////////////////


int query5()
{
    AssocItr* modI; // iterator for traversing AllModules collection
    Module* modH;  
    AssocItr*	baseI;  // iterator for iterating over all base
			// assemblies in a module
    BaseAssembly* baseH;
    AssocItr*	compI;  // iterator for iterating over all composite parts
			// in a base assembly
    CompositePart* compH;

    // scan all modules, scanning all of the base assemblies in each one,
    // scanning all of the composite parts that they use, looking for base
    // assemblies that use more recently dated composite parts

    int count = 0;
    modI = new AssocItr(&AllModules);
    modH = (Module*) modI->next();
    while (modH != NULL)
    {
	// for each module iterate over its base assemblies
	// this assumes that we maintain a collection of base asemblies
	// in each module
    	baseI = new AssocItr(modH->allBases); 
    	baseH = (BaseAssembly*) baseI->next();
    	while (baseH != NULL)
	{
            if (debugMode) {
                printf("In Query5, processing base assembly %d, buildDate = %d\n",
		            baseH->id, baseH->buildDate);
            }
    	    compI = new AssocItr(baseH->componentsPriv); 
    	    compH = (CompositePart*) compI->next();
	    while (compH != NULL)
	    {
                if (debugMode) {
                    printf("            [Checking composite part %d, buildDate = %d]\n",
		                        compH->id, compH->buildDate);
                }
		if (compH->buildDate > baseH->buildDate) {
	            baseH->DoNothing();
		    count++;
	        }
    	        compH = (CompositePart*) compI->next();
	    }
	    delete compI;
    	    baseH = (BaseAssembly*) baseI->next();
	}
	delete baseI;
        modH = (Module*) modI->next();
    }
    delete modI;
    return count;
}
